@@ -21,21 +21,50 @@ import difflib
 import jinja2
 import os
 import re
+import signal
 import subprocess
 import sys
 import time
 import tempfile
+import urlparse
 import xmlrpclib
 import yaml
+import zmq
 
-from lava_tool.authtoken import AuthenticatingServerProxy, KeyringAuthBackend
+from lava_tool.authtoken import (
+    split_xmlrpc_url,
+    get_server_url,
+    AuthenticatingServerProxy,
+    KeyringAuthBackend
+)
 from lava.tool.command import Command, CommandGroup
 from lava.tool.errors import CommandError
 from lava_dashboard_tool.commands import DataSetRenderer, _get_pretty_renderer
 from lava_scheduler_tool.scheduler import (
     devicedictionary_to_jinja2,
-    jinja2_to_devicedictionary
+    jinja2_to_devicedictionary,
+    validate_urls
 )
+from zmq.utils.strtypes import b, u
+
+
+class Timeout():
+    """ Timeout error class with ALARM signal. Accepts time in seconds. """
+    class TimeoutError(Exception):
+        pass
+
+    def __init__(self, sec):
+        self.sec = sec
+
+    def __enter__(self):
+        signal.signal(signal.SIGALRM, self.timeout_raise)
+        signal.alarm(self.sec)
+
+    def __exit__(self, *args):
+        signal.alarm(0)
+
+    def timeout_raise(self, *args):
+        raise Timeout.TimeoutError()
 
 
 class scheduler(CommandGroup):
@@ -56,31 +85,47 @@ class submit_job(Command):
         super(submit_job, cls).register_arguments(parser)
         parser.add_argument("SERVER")
         parser.add_argument("JSON_FILE")
-        parser.add_argument("--block",
-                            action="store_true",
-                            help="Blocks until the job gets executed")
+        parser.add_argument(
+            "--block",
+            action="store_true",
+            help="Blocks until the job gets executed. Deprecated. This option adds unnecessary load on the server. Please use wait-job-events command instead.")
 
     def invoke(self):
+        auth_backend = KeyringAuthBackend()
         server = AuthenticatingServerProxy(
-            self.args.SERVER, auth_backend=KeyringAuthBackend())
+            self.args.SERVER, auth_backend=auth_backend)
         if not os.path.exists(self.args.JSON_FILE):
             raise CommandError("No such file: %s" % self.args.JSON_FILE)
         with open(self.args.JSON_FILE, 'rb') as stream:
             command_text = stream.read()
         try:
+            validate_urls(command_text)
             job_ids = server.scheduler.submit_job(command_text)
         except xmlrpclib.Fault, e:
             raise CommandError(str(e))
         else:
+            # Need to get the host in case of shortcuts usage.
+            host = get_server_url(auth_backend, self.args.SERVER)
             if isinstance(job_ids, list):
-                print "submitted as job ids:"
+                print "submitted as jobs:"
                 for job_id in job_ids:
-                    print " -", job_id
+                    job = server.scheduler.job_details(job_id)
+                    if 'absolute_url' in job:
+                        print urlparse.urljoin(host, job["absolute_url"])
+                    else:
+                        print " -", job_id
             else:
-                print "submitted as job id:", job_ids
+                job = server.scheduler.job_details(job_ids)
+                if 'absolute_url' in job:
+                    print "submitted as job:", urlparse.urljoin(
+                        host, job["absolute_url"])
+                else:
+                    print "submitted as job:", job_ids
                 job_ids = [job_ids]
 
         if self.args.block:
+            print('')
+            print('This kind of polling is deprecated and will be removed in the next release. Please use "wait-for-job" command.')
             print('')
             print('Waiting for the job to run ')
             print('. = job waiting in the queue')
@@ -108,14 +153,20 @@ class resubmit_job(Command):
         parser.add_argument("JOB_ID")
 
     def invoke(self):
+        auth_backend = KeyringAuthBackend()
         server = AuthenticatingServerProxy(
-            self.args.SERVER, auth_backend=KeyringAuthBackend())
+            self.args.SERVER, auth_backend=auth_backend)
         try:
             job_id = server.scheduler.resubmit_job(self.args.JOB_ID)
+            # Need to get the host in case of shortcuts usage.
+            host = get_server_url(auth_backend, self.args.SERVER)
+
         except xmlrpclib.Fault, e:
             raise CommandError(str(e))
         else:
-            print "resubmitted as job id:", job_id
+            job = server.scheduler.job_details(job_id)
+            print "resubmitted as job:", urlparse.urljoin(
+                host, job["absolute_url"])
 
 
 class cancel_job(Command):
@@ -566,3 +617,156 @@ class device_dictionary(Command):
                             help="Load a jinja2 file to update the device dictionary")
         parser.add_argument("--export", "-e", action='store_true',
                             help="Export the device dictionary for this device as jinja2")
+
+
+class wait_job_events(Command):
+    """
+    Wait for job to finish and return job details.
+    """
+
+    FINISHED_JOB_STATUS = ["Complete", "Incomplete", "Canceled"]
+    PUBLISHER_PROTOCOL = "tcp://"
+    PUBLISHER_PORT = 5500
+
+    @classmethod
+    def register_arguments(cls, parser):
+        super(wait_job_events, cls).register_arguments(parser)
+        parser.add_argument("SERVER")
+        action = parser.add_mutually_exclusive_group(required=True)
+        action.add_argument("--job-definition",
+                            help="Definition file path of the job that will be submitted.")
+        action.add_argument("--job-id",
+                            type=int,
+                            help="ID of the job to wait for. If this argument is provided, this command will not submit new job.")
+        parser.add_argument("--port", "-p",
+                            type=int,
+                            help="Alternative publisher port to use")
+        parser.add_argument("--timeout",
+                            type=int,
+                            default=0,
+                            help="Time to wait until the job is finished. Default will wait until manually canceled.")
+        parser.add_argument(
+            "--quiet", "-q",
+            action="store_true",
+            help="Do not print out interim messages (when job becomes Submitted or Running) but only when the job is finished.")
+
+    def invoke(self):
+        auth_backend = KeyringAuthBackend()
+        server = AuthenticatingServerProxy(
+            self.args.SERVER, auth_backend=auth_backend)
+
+        # Need to get the host in case of shortcuts usage.
+        parsed_host = urlparse.urlparse(
+            get_server_url(auth_backend, self.args.SERVER))
+        host = parsed_host.netloc
+
+        if self.args.job_id:
+            job_ids = [self.args.job_id]
+            try:
+                job_details = server.scheduler.job_details(self.args.job_id)
+
+                if job_details["status"] in self.FINISHED_JOB_STATUS:
+                    print "Job %s already finished with status %s." % (
+                        self.args.job_id, job_details["status"])
+                    return
+
+            except xmlrpclib.Fault as e:
+                print "Job details did not found specified job. Please make sure job-id exists or that this job is about to get submitted."
+
+        else:
+            if not os.path.exists(self.args.job_definition):
+                raise CommandError(
+                    "No such file: %s" % self.args.job_definition)
+            try:
+                with open(self.args.job_definition, 'rb') as stream:
+                    command_text = stream.read()
+            except IOError as e:
+                print >> sys.stderr, e
+                return -1
+            try:
+                validate_urls(command_text)
+                job_ids = server.scheduler.submit_job(command_text)
+                print "submitted as job id(s):", job_ids
+                if not isinstance(job_ids, list):
+                    job_ids = [job_ids]
+
+            except xmlrpclib.Fault, e:
+                raise CommandError(str(e))
+
+        port = self.args.port if self.args.port else self.PUBLISHER_PORT
+        event_socket = "%s%s:%s" % (self.PUBLISHER_PROTOCOL, host, port)
+
+        try:
+            with Timeout(self.args.timeout):
+                context = zmq.Context.instance()
+                sock = context.socket(zmq.SUB)
+
+                sock.setsockopt(zmq.SUBSCRIBE, b"")
+                sock.connect(event_socket)
+
+                print >> sys.stdout, "Now waiting for job events..."
+                while True:
+                    msg = sock.recv_multipart()
+                    try:
+                        (topic, uuid, dt, username, data) = msg[:]
+                    except IndexError:
+                        # Droping invalid message
+                        continue
+
+                    data = yaml.safe_load(data)
+                    if "job" in data:
+                        # Filter out device events.
+                        if "submitter" in data:
+                            if data["job"] in job_ids:
+                                if data["status"] in self.FINISHED_JOB_STATUS:
+                                    print data
+                                    break
+                                elif not self.args.quiet:
+                                    print data
+
+        except zmq.error.ZMQError as e:
+            print >> sys.stderr, "wait_job_events() error: %s" % e
+            return -1
+        except KeyboardInterrupt:
+            print >> sys.stderr, "wait_job_events() cancelled"
+            return -1
+        except Timeout.TimeoutError:
+            print >> sys.stderr, "wait_job_events() timed out after %s seconds." % self.args.timeout
+            try:
+                for job_id in job_ids:
+                    job_details = server.scheduler.job_details(job_id)
+                    print >> sys.stdout, "Job %s currently has status %s." % (
+                        job_id, job_details["status"])
+                    print >> sys.stdout, "You can check the job details with 'lava-tool job-details' command."
+                return 1
+            except xmlrpclib.Fault as e:
+                # Keep silent, message about non-existing job is already
+                # printed during first job_details() call.
+                return -1
+
+        return 0
+
+
+class validate_pipeline_devices(Command):
+    """
+    Validate pipeline devices on the specified server.
+    """
+
+    @classmethod
+    def register_arguments(cls, parser):
+        super(validate_pipeline_devices, cls).register_arguments(parser)
+        parser.add_argument("SERVER")
+        parser.add_argument("--name",
+                            default="",
+                            help="Hostname of the device or device_type name.")
+
+    def invoke(self):
+        auth_backend = KeyringAuthBackend()
+        server = AuthenticatingServerProxy(
+            self.args.SERVER, auth_backend=auth_backend)
+
+        try:
+            print server.scheduler.validate_pipeline_devices(
+                self.args.name)
+        except xmlrpclib.Fault, e:
+            raise CommandError(str(e))
